package more.ssaqms.xiamen.model.report;

import lombok.Data;
import lombok.ToString;
import lombok.experimental.Accessors;
import more.ssaqms.xiamen.model.statistics.AqiData;
import more.ssaqms.xiamen.model.statistics.PollutantData;
import more.ssaqms.xiamen.util.CalculationUtils;

import java.util.*;

/**
 * @author purpose
 * @version 1.0
 * @date 2023/5/6 3:32 下午
 */
@Data
public class ReportAQILinGangData {

  // 按日统计的AQI数据列表
  private List<AqiData> aqiDataList = null;

  // 根据AQI等级统计的天数
  private List<AQILevelDayStatistics> aqiLevelDayStatistics = null;

  // 各个AQI等级的日期
  private List<String>[] aqiLevelDates = null;

  // 优良天数和占比:
  // 下标0为优天数;
  // 下标1为良好天数;
  // 下标2为污染天数;
  // 下标3为优良天数占比;
  // 下标4为优天数占比;
  // 下标5为良好天数占有比;
  // 下标6为污染天数占比
  private float[] goodDayPercentage = null;

  // 获取有数据的总天数
  public int getTotalDays() {
    return this.aqiDataList.size();
  }

  // 首要污染物列表, 如果为null, 代表无首要污染物
  private List<String> primaryPollutants = null;

  // O₃_8H平均值
  private Float o38hAvg = null;

  // PM₂.₅平均值
  private Float pm25Avg = null;

  // PM₁₀最大值最小值
  private Float[] pm10ValueRange = null;

  // PM₂.₅最大值最小值
  private Float[] pm25ValueRange = null;

  // SO₂最大值最小值
  private Float[] so2ValueRange = null;

  // CO最大值最小值
  private Float[] coValueRange = null;

  // NO₂最大值最小值
  private Float[] no2ValueRange = null;

  // O₃_8H最大值最小值
  private Float[] o38hValueRange = null;

  /**
   * 构造函数
   * @param aqiDataList
   */
  public ReportAQILinGangData(List<AqiData> aqiDataList) {
    this.aqiDataList = aqiDataList;
    this.statisticsAQILevelDays();
    this.setPrimaryPollutant();
    this.setAvgData();
  }

  private void statisticsAQILevelDays() {
    final int AQI_LEVEL_LENGTH = AQI_LEVEL_NAMES.length;

    int[] aqiLevelDays = new int[AQI_LEVEL_LENGTH];
    List<String>[] aqiLevelDates = new List[AQI_LEVEL_LENGTH];
    for (int i = 0; i < AQI_LEVEL_LENGTH; i++) {
      aqiLevelDays[i] = 0;
      aqiLevelDates[i] = new ArrayList<>();
    }

    final List<AqiData> aqiDataList = this.aqiDataList;
    Calendar ca = Calendar.getInstance();

    for (AqiData aqiData : aqiDataList) {
      int index = aqiData.getGrade() - 1;
      Date dataTime = aqiData.getDataTime();
      ca.setTime(dataTime);

      aqiLevelDays[index]++;
      aqiLevelDates[index].add(ca.get(Calendar.DAY_OF_MONTH) + "日");
    }

    List<AQILevelDayStatistics> aqiLevelDayStatistics = new ArrayList<>();
    float[] goodDayPercentage = new float[]{0, 0, 0, 0, 0, 0, 0};

    for (int i = 0; i < AQI_LEVEL_LENGTH; i++) {
      int days = aqiLevelDays[i];

      if (days > 0) {
        aqiLevelDayStatistics.add(new AQILevelDayStatistics(i + 1, AQI_LEVEL_NAMES[i], days, AQI_LEVEL_COLORS[i]));

        if (i < 2) {
          goodDayPercentage[i] = days;
        } else {
          goodDayPercentage[2] += days;
        }
      }
    }

    int totalDays = this.getTotalDays();
    goodDayPercentage[3] = CalculationUtils.digit2((float) (goodDayPercentage[0] + goodDayPercentage[1]) / totalDays * 100, 2);
    goodDayPercentage[4] = CalculationUtils.digit2((float) goodDayPercentage[0] / totalDays * 100, 2);
    goodDayPercentage[5] = CalculationUtils.digit2((float) goodDayPercentage[1] / totalDays * 100, 2);
    goodDayPercentage[6] = CalculationUtils.digit2((float) goodDayPercentage[2] / totalDays * 100, 2);

    this.goodDayPercentage = goodDayPercentage;
    this.aqiLevelDayStatistics = aqiLevelDayStatistics;
    this.aqiLevelDates = aqiLevelDates;
  }

  private void setPrimaryPollutant() {
    float[] goodDayPercentage = this.goodDayPercentage;

    if (goodDayPercentage[1] == 0 && goodDayPercentage[2] == 0) {
      return;
    }

    List<AqiData> aqiDataList = this.aqiDataList;
    Set<String> pollutantSet = new HashSet<>();

    for (AqiData aqiData : aqiDataList) {
      String primaryPollutant = aqiData.getPrimaryPollutant();

      if (primaryPollutant == null || primaryPollutant.equals("无")) {
        continue;
      }

      String[] primaryPollutants = primaryPollutant.split(",");

      for (String pollutantName : primaryPollutants) {
        String convertPollutantName = FACTOR_NAME_MAPPING.get(pollutantName);
        pollutantSet.add((convertPollutantName == null ? pollutantName : convertPollutantName));
      }
    }

    this.primaryPollutants = new ArrayList<>(pollutantSet);
  }

  private void setAvgData() {
    List<AqiData> aqiDataList = this.aqiDataList;

    float[] o38h = new float[]{0f, 0f};
    float[] pm25 = new float[]{0f, 0f};

    Float[] pm10ValueRange = new Float[]{null, null};
    Float[] pm25ValueRange = new Float[]{null, null};
    Float[] so2ValueRange = new Float[]{null, null};
    Float[] coValueRange = new Float[]{null, null};
    Float[] no2ValueRange = new Float[]{null, null};
    Float[] o38hValueRange = new Float[]{null, null};

    for (AqiData aqiData : aqiDataList) {
      List<PollutantData> pollutantDatas = aqiData.getPollutantData();

      if (pollutantDatas == null) {
        continue;
      }

      for (PollutantData pollutantData : pollutantDatas) {
        Float value = pollutantData.getValue();

        if (value != null) {
          switch (pollutantData.getFactorId()) {
            case "1001":
              if (so2ValueRange[0] == null) {
                so2ValueRange[0] = so2ValueRange[1] = value;
              } else if (value < so2ValueRange[0]) {
                so2ValueRange[0] = value;
              } else if (value > so2ValueRange[1]) {
                so2ValueRange[1] = value;
              }
              break;
            case "1003":
              if (no2ValueRange[0] == null) {
                no2ValueRange[0] = no2ValueRange[1] = value;
              } else if (value < no2ValueRange[0]) {
                no2ValueRange[0] = value;
              } else if (value > no2ValueRange[1]) {
                no2ValueRange[1] = value;
              }
              break;
            case "1005":
              if (coValueRange[0] == null) {
                coValueRange[0] = coValueRange[1] = value;
              } else if (value < coValueRange[0]) {
                coValueRange[0] = value;
              } else if (value > coValueRange[1]) {
                coValueRange[1] = value;
              }
              break;
            case "1006":
              if (pollutantData.getRootFactorId().equals("H8O3")) {
                o38h[0] += value;
                o38h[1]++;

                if (o38hValueRange[0] == null) {
                  o38hValueRange[0] = o38hValueRange[1] = value;
                } else if (value < o38hValueRange[0]) {
                  o38hValueRange[0] = value;
                } else if (value > o38hValueRange[1]) {
                  o38hValueRange[1] = value;
                }
              }
              break;
            case "1007":
              if (pm10ValueRange[0] == null) {
                pm10ValueRange[0] = pm10ValueRange[1] = value;
              } else if (value < pm10ValueRange[0]) {
                pm10ValueRange[0] = value;
              } else if (value > pm10ValueRange[1]) {
                pm10ValueRange[1] = value;
              }
              break;
            case "1008":
              pm25[0] += value;
              pm25[1]++;

              if (pm25ValueRange[0] == null) {
                pm25ValueRange[0] = pm25ValueRange[1] = value;
              } else if (value < pm25ValueRange[0]) {
                pm25ValueRange[0] = value;
              } else if (value > pm25ValueRange[1]) {
                pm25ValueRange[1] = value;
              }
              break;
          }
        }
      }
    }

    if (o38h[1] > 0) {
      this.o38hAvg = CalculationUtils.digit2(o38h[0] / o38h[1], 2);
    }

    if (pm25[1] > 0) {
      this.pm25Avg = CalculationUtils.digit2(pm25[0] / pm25[1], 2);
    }

    this.pm10ValueRange = new Float[]{pm10ValueRange[0] == null ? null : CalculationUtils.digit2(pm10ValueRange[0], 2), pm10ValueRange[1] == null ? null : CalculationUtils.digit2(pm10ValueRange[1], 2)};
    this.pm25ValueRange = new Float[]{pm25ValueRange[0] == null ? null : CalculationUtils.digit2(pm25ValueRange[0], 2), pm25ValueRange[1] == null ? null : CalculationUtils.digit2(pm25ValueRange[1], 2)};
    this.so2ValueRange = new Float[]{so2ValueRange[0] == null ? null : CalculationUtils.digit2(so2ValueRange[0], 2), so2ValueRange[1] == null ? null : CalculationUtils.digit2(so2ValueRange[1], 2)};
    this.coValueRange = new Float[]{coValueRange[0] == null ? null : CalculationUtils.digit2(coValueRange[0], 2), coValueRange[1] == null ? null : CalculationUtils.digit2(coValueRange[1], 2)};
    this.no2ValueRange = new Float[]{no2ValueRange[0] == null ? null : CalculationUtils.digit2(no2ValueRange[0], 2), no2ValueRange[1] == null ? null : CalculationUtils.digit2(no2ValueRange[1], 2)};
    this.o38hValueRange = new Float[]{o38hValueRange[0] == null ? null : CalculationUtils.digit2(o38hValueRange[0], 2), o38hValueRange[1] == null ? null : CalculationUtils.digit2(o38hValueRange[1], 2)};
  }


  @Data
  @ToString
  @Accessors(chain = true)
  public static final class ItemStyle {
    String color;

    public ItemStyle(String color) {
      this.color = color;
    }
  }

  @Data
  @ToString
  @Accessors(chain = true)
  public static final class AQILevelDayStatistics {
    private int level;

    private String name;

    private int value;

    private ItemStyle itemStyle;

    public AQILevelDayStatistics(int level, String name, int value, String color) {
      this.level = level;
      this.name = name;
      this.value = value;
      this.itemStyle = new ItemStyle(color);
    }
  }

  public static final String[] AQI_LEVEL_NAMES = new String[]{
      "优",
      "良",
      "轻度污染",
      "中度污染",
      "重度污染",
      "严重污染"
  };

  public static final String[] AQI_LEVEL_COLORS = new String[]{
      "rgb(0, 228, 0)",
      "rgb(255, 255, 0)",
      "rgb(255, 126, 0)",
      "rgb(255, 0, 0)",
      "rgb(153, 0, 76)",
      "rgb(126, 0, 35)"
  };

  public static final Map<String, String> FACTOR_NAME_MAPPING = new HashMap() {{
    put("SO2", "SO₂");
    put("NOX", "NO᙮");
    put("NO2", "NO₂");
    put("O3", "O₃");
    put("PM10", "PM₁₀");
    put("PM2.5", "PM₂.₅");
    put("H8O3", "O₃");
  }};
}
