package more.ssaqms.xiamen.model.report;

import lombok.Data;
import lombok.ToString;
import lombok.experimental.Accessors;
import more.ssaqms.xiamen.model.data.HistoricalData;
import more.ssaqms.xiamen.model.factor.Factor;
import more.ssaqms.xiamen.model.factor.FactorGroup;
import more.ssaqms.xiamen.model.site.ReportSiteDevice;
import more.ssaqms.xiamen.model.site.Site;
import more.ssaqms.xiamen.util.CalculationUtils;

import java.text.SimpleDateFormat;
import java.util.*;


/**
 * @author purpose
 * @version 1.0
 * @date 2023/5/9 8:26 上午
 */
@Data
@ToString
@Accessors(chain = true)
public class ReportMonthVOCsLinGangData {

  private final TimeRange timeRange;

  // 站点
  private List<Site> sites;

  // 所有参与统计的站点VOCs均值
  private Float[] vocsValueRange;

  // top VOCs物质列表
  private List<SiteVOCsTopData> siteVOCsTopDataList;

  // VOCs主要污染物
  private List<String> primaryPollutants;

  // 站点VOCs物质、分类、总量列表
  private List<SiteRangeAvgData> siteRangeAvgDataList;

  private int totalHours = 0;
  private int totalDays = 0;

  public ReportMonthVOCsLinGangData(
    TimeRange timeRange,
    List<String> siteIds,
    List<ReportSiteDevice> siteDevices,
    BasicInformation information,
    List<HistoricalData> dataList,
    ChartData nmhcChartData
  ) {
    this.timeRange = timeRange;
    this.setTotalHours(timeRange)
      .setTotalDays(timeRange)
      .setSites(siteIds, information)
      .statisticSiteAvg(information, siteDevices, dataList, nmhcChartData);
  }

  private ReportMonthVOCsLinGangData setTotalHours(TimeRange timeRange) {
    long diff = timeRange.getEndTime().getTime() - timeRange.getBeginTime().getTime();
    this.totalHours = (int) (diff / 1000 / 60 / 60 + 1);
    return this;
  }

  private ReportMonthVOCsLinGangData setTotalDays(TimeRange timeRange) {
    long diff = timeRange.getEndTime().getTime() - timeRange.getBeginTime().getTime();
    this.totalDays = (int) ((diff / 1000 / 60 / 60 + 1) / 24);
    return this;
  }


  private ReportMonthVOCsLinGangData setSites(
    List<String> siteIds,
    BasicInformation information
  ) {
    Map<String, Site> siteMap = information.getSiteMap();
    Map<String, Integer> siteIndexMap = new HashMap<>();

    siteMap.forEach((siteId, site) -> {
      siteIndexMap.put(siteId, siteIndexMap.size());
    });

    List<Site> sites = new ArrayList();

    for (String siteId : siteIds) {
      sites.add(siteMap.get(siteId));
    }

    sites.sort(new Comparator<Site>() {
      @Override
      public int compare(Site o1, Site o2) {
        return siteIndexMap.get(o1.getId()) - siteIndexMap.get(o2.getId());
      }
    });

    this.sites = sites;

    return this;
  }

  private ReportMonthVOCsLinGangData statisticSiteAvg(
    BasicInformation information,
    List<ReportSiteDevice> siteDevices,
    List<HistoricalData> dataList,
    ChartData nmhcChartData
  ) {
    Map<String, Map<String, ReportMonthAvgStatistic>> siteSumMap = this.getSiteSumMap(siteDevices);
    Map<String, Map<String, Object[]>> siteFactorMaxValueMap = this.getSiteFactorMaxValueMap(siteDevices);
    Map<String, Map<Long, Map<String, ReportMonthAvgStatistic>>> siteDailySumMap = this.getSiteDailySumMap(siteDevices);
    Map<String, Map<Long, Float[]>> siteNMHCAvgDataMap = this.getSiteNMHCAvgDataMap(information, nmhcChartData);
    this.setSiteSum(siteSumMap, siteFactorMaxValueMap, siteDailySumMap, dataList, information)
      .setSiteVOCsTopDataList(siteSumMap)
      .setDailySiteVOCsData(siteDailySumMap, siteFactorMaxValueMap, siteNMHCAvgDataMap, information);
    return this;
  }

  private Map<String, Map<Long, Float[]>> getSiteNMHCAvgDataMap(BasicInformation information, ChartData nmhcChartData) {
    Factor factor = information.getFactorMap().get("1026");
    String stdUnit = factor.getStdUnit();
    String displayUnit = factor.getDisplayUnit();
    Float molecular = factor.getMolecular();

    List<HistoricalData> dataList = nmhcChartData.getDataList();
    Map<String, Map<Long, Float[]>> siteNMHCAvgDataMap = new HashMap();

    dataList.forEach(data -> {
      Float value = data.getValue();
      if (value == null) {
        return;
      }

      String siteId = data.getSiteId();
      Map<Long, Float[]> dataTimeMap = siteNMHCAvgDataMap.get(siteId);
      if (dataTimeMap == null) {
        siteNMHCAvgDataMap.put(siteId, dataTimeMap = new LinkedHashMap());
      }

      Date dataTime = data.getDataTime();
      long ts = dataTime.getTime();
      long mod = ts % 86400000;
      if (mod >= 57600000) {
        mod -= 57600000;
      } else {
        mod += 28800000;
      }
      long newTs = ts - mod;

      Float[] avgData = dataTimeMap.get(newTs);
      if (avgData == null) {
        dataTimeMap.put(newTs, avgData = new Float[]{0f, 0f});
      }

      avgData[0] += CalculationUtils.UnitConverter(value, stdUnit, displayUnit, molecular);
      avgData[1]++;
    });

    return siteNMHCAvgDataMap;
  }

  private Map<String, Map<String, ReportMonthAvgStatistic>> getSiteSumMap(List<ReportSiteDevice> siteDevices) {
    Map<String, Map<String, ReportMonthAvgStatistic>> siteSumMap = new HashMap<>();

    List<Site> sites = this.sites;
    for (Site site : sites) {
      siteSumMap.put(site.getId(), new HashMap<String, ReportMonthAvgStatistic>());
    }

    for (ReportSiteDevice siteDevice : siteDevices) {
      Map<String, ReportMonthAvgStatistic> factorAvgStatisticMap = siteSumMap.get(siteDevice.getSiteId());

      siteDevice.getFactorMap().forEach((factorId, factor) -> {
        factorAvgStatisticMap.put(factorId, new ReportMonthAvgStatistic(factor));
      });
    }

    return siteSumMap;
  }

  private Map<String, Map<String, Object[]>> getSiteFactorMaxValueMap(List<ReportSiteDevice> siteDevices) {
    Map<String, Map<String, Object[]>> siteFactorMaxValueMap = new HashMap<>();

    List<Site> sites = this.sites;
    for (Site site : sites) {
      siteFactorMaxValueMap.put(site.getId(), new HashMap());
    }

    for (ReportSiteDevice siteDevice : siteDevices) {
      Map<String, Object[]> factorMaxValueMap = siteFactorMaxValueMap.get(siteDevice.getSiteId());
    }

    return siteFactorMaxValueMap;
  }

  private Map<String, Map<Long, Map<String, ReportMonthAvgStatistic>>> getSiteDailySumMap(List<ReportSiteDevice> siteDevices) {
    Map<String, Map<Long, Map<String, ReportMonthAvgStatistic>>> siteSumMap = new LinkedHashMap();

    List<Site> sites = this.sites;
    for (Site site : sites) {
      siteSumMap.put(site.getId(), new LinkedHashMap<>());
    }

    return siteSumMap;
  }

  private ReportMonthVOCsLinGangData setSiteSum(
    Map<String, Map<String, ReportMonthAvgStatistic>> siteSumMap,
    Map<String, Map<String, Object[]>> siteFactorMaxValueMap,
    Map<String, Map<Long, Map<String, ReportMonthAvgStatistic>>> siteDailySumMap,
    List<HistoricalData> dataList,
    BasicInformation information
  ) {
    Map<String, Factor> factorMap = information.getFactorMap();

    SimpleDateFormat dateFormat = new SimpleDateFormat("MM-dd HH");

    for (HistoricalData data : dataList) {
      String siteId = data.getSiteId();
      Map<String, ReportMonthAvgStatistic> factorReportMonthAvgStatisticMap = siteSumMap.get(siteId);

      if (factorReportMonthAvgStatisticMap == null) {
        continue;
      }

      String factorId = data.getFactorId();
      ReportMonthAvgStatistic reportMonthAvgStatistic = factorReportMonthAvgStatisticMap.get(factorId);

      if (reportMonthAvgStatistic == null) {
        continue;
      }

      Float value = data.getValue();
      reportMonthAvgStatistic.setSum(value);

      Map<Long, Map<String, ReportMonthAvgStatistic>> factorDataMap = siteDailySumMap.get(siteId);

      Date dataTime = data.getDataTime();
      long ts = dataTime.getTime();

      long mod = ts % 86400000;
      if (mod >= 57600000) {
        mod -= 57600000;
      } else {
        mod += 28800000;
      }
      long newTs = ts - mod;

      Map<String, ReportMonthAvgStatistic> tsAvgStatisticMap = factorDataMap.get(newTs);

      if (tsAvgStatisticMap == null) {
        factorDataMap.put(newTs, tsAvgStatisticMap = new HashMap<>());
      }

      ReportMonthAvgStatistic reportDailyAvgStatistic = tsAvgStatisticMap.get(factorId);

      if (reportDailyAvgStatistic == null) {
        tsAvgStatisticMap.put(factorId, reportDailyAvgStatistic = new ReportMonthAvgStatistic(reportMonthAvgStatistic.getFactor()));
      }

      reportDailyAvgStatistic.setSum(value);

      // 设置最大值和日期
      Map<String, Object[]> factorMaxValueMap = siteFactorMaxValueMap.get(siteId);
      if (factorMaxValueMap != null) {
        Object[] maxValueInfo = factorMaxValueMap.get(factorId);
        if (maxValueInfo == null) {
          factorMaxValueMap.put(factorId, maxValueInfo = new Object[]{-999999f, ""});
        }

        Factor factor = factorMap.get(factorId);
        value = CalculationUtils.UnitConverter(value, factor.getStdUnit(), factor.getDisplayUnit(), factor.getMolecular());

        if (value > (Float) maxValueInfo[0]) {
          maxValueInfo[0] = value;
          maxValueInfo[1] = dateFormat.format(dataTime);
        }
      }
    }

    return this;
  }

  private ReportMonthVOCsLinGangData setSiteVOCsTopDataList(Map<String, Map<String, ReportMonthAvgStatistic>> siteSumMap) {
    List<Float> vocsValues = new ArrayList();
    List<SiteVOCsTopData> topDataList = new ArrayList();
    Set<String> primaryPollutantSet = new HashSet<>();

    List<Site> sites = this.sites;
    int totalHours = this.totalHours;
    int totalDays = this.totalDays;

    for (Site site : sites) {
      String siteId = site.getId();
      Map<String, ReportMonthAvgStatistic> factorAvgStatisticMap = siteSumMap.get(siteId);

      if (factorAvgStatisticMap.size() == 0) {
        continue;
      }

      String siteName = site.getName();
      List<SiteVOCsTopData> avgDataList = new ArrayList();
      Float[] vocsSum = new Float[]{null};

      factorAvgStatisticMap.forEach((factorId, avgStatistic) -> {
        int count = avgStatistic.getCount();

        float percentage = (float) count / totalHours * 100;

        if (percentage <= 30) {
          return;
        }

        Factor factor = avgStatistic.getFactor();

        float avg = CalculationUtils.UnitConverter(avgStatistic.getSum(), factor.getStdUnit(), factor.getDisplayUnit(), factor.getMolecular()) / count;
        SiteVOCsTopData siteVOCsTopData = new SiteVOCsTopData(siteName, avgStatistic.getFactor().getDisplayName(), CalculationUtils.digit2(avg, 2));
        avgDataList.add(siteVOCsTopData);

        if (vocsSum[0] == null) {
          vocsSum[0] = 0f;
        }
        vocsSum[0] += avg;
      });

      Float vocsValue = vocsSum[0] == null ? null : CalculationUtils.digit2(vocsSum[0], 2);
      topDataList.add(new SiteVOCsTopData(siteName, "VOCs总量", vocsValue));
      vocsValues.add(vocsValue);
      avgDataList.sort(new Comparator<SiteVOCsTopData>() {
        @Override
        public int compare(SiteVOCsTopData o1, SiteVOCsTopData o2) {
          Float v1 = o1.getValue();
          Float v2 = o2.getValue();
          return v1 == null || v2 == null || v1 < v2 ? 1 : -1;
        }
      });

      int len = Math.min(3, avgDataList.size());

      for (int i = 0; i < len; i++) {
        SiteVOCsTopData siteVOCsTopData = avgDataList.get(i);
        topDataList.add(siteVOCsTopData);
        primaryPollutantSet.add(siteVOCsTopData.getFactorName());
      }
    }

    List<Float> newVocsValueRange = new ArrayList<>();
    vocsValues.forEach(vocsValue -> {
      if (vocsValue != null) {
        newVocsValueRange.add(vocsValue);
      }
    });

    int size = newVocsValueRange.size();

    if (size > 0) {
      newVocsValueRange.sort(new Comparator<Float>() {
        @Override
        public int compare(Float o1, Float o2) {
          return o1 > o2 ? 1 : o1 < o2 ? -1 : 0;
        }
      });

      this.vocsValueRange = new Float[]{newVocsValueRange.get(0), newVocsValueRange.get(size - 1)};
    }

    this.primaryPollutants = new ArrayList<>(primaryPollutantSet);
    this.siteVOCsTopDataList = topDataList;

    return this;
  }

  private ReportMonthVOCsLinGangData setDailySiteVOCsData(
    Map<String, Map<Long, Map<String, ReportMonthAvgStatistic>>> siteDailySumMap,
    Map<String, Map<String, Object[]>> siteFactorMaxValueMap,
    Map<String, Map<Long, Float[]>> siteNMHCAvgDataMap,
    BasicInformation information
  ) {
    Map<String, Site> siteMap = information.getSiteMap();
    Map<String, FactorGroup> factorCategoryMap = information.getFactorCategoryMap();
    Map<String, Factor> factorMap = information.getFactorMap();
    List<SiteRangeAvgData> list = new ArrayList<>();

    Map<String, Float> ofpFactorMap = this.getOfpFactorMap(factorMap);
    Map<String, Float> soapFactorMap = this.getSoapFactorMap(factorMap);

    siteDailySumMap.forEach((siteId, factorDailyMap) -> {
      Site site = siteMap.get(siteId);
      list.add(
        new SiteRangeAvgData(
          site.getName(),
          site.getType(),
          factorDailyMap,
          factorMap,
          factorCategoryMap,
          siteFactorMaxValueMap.get(siteId),
          siteNMHCAvgDataMap.get(siteId),
          ofpFactorMap,
          soapFactorMap
        ));
    });

    this.siteRangeAvgDataList = list;

    return this;
  }

  private Map<String, Float> getOfpFactorMap(Map<String, Factor> factorMap) {
    Map<String, Float> ofpFactorMap = new HashMap<>();

    factorMap.forEach((factorId, factor) -> {
      Float mir = factor.getMir();

      if (!Objects.isNull(mir)) {
        ofpFactorMap.put(factorId, mir);
      }
    });

    return ofpFactorMap;
  }

  private Map<String, Float> getSoapFactorMap(Map<String, Factor> factorMap) {
    Map<String, Float> soapFactorMap = new HashMap<>();

    factorMap.forEach((factorId, factor) -> {
      Float fvocr = factor.getFvocr();

      if (!Objects.isNull(fvocr)) {
        soapFactorMap.put(factorId, 1 / ((1 - fvocr) / factor.getFac()));
      }
    });

    return soapFactorMap;
  }

  @Data
  @ToString
  @Accessors(chain = true)
  public static final class SiteVOCsTopData {
    private String siteName;

    private String factorName;

    private Float value;

    public SiteVOCsTopData(String siteName, String factorName, Float value) {
      this.siteName = siteName;
      this.factorName = factorName;
      this.value = value;
    }
  }

  @Data
  @ToString
  @Accessors(chain = true)
  public static final class SiteRangeAvgData {
    // 站点名称
    private String siteName;

    private Long siteType;

    // VOCs浓度范围
    private Float[] range;

    // VOCs浓度均值
    private Float avg;

    // 分类饼图数据
    private List<PieData> pieDataList;

    // VOCs浓度(单位μg/m³)超过1.2(如果没有超过1.2, 至少五个物质)的列表
    private List<Object[]> topVocFactorTableData;

    // 站点VOCs月均值排名
    private List<PieData> factorAvgList;

    // nmhc日均值图表数据
    private List<HistoricalData> nmhcAvgDailyDataList = new ArrayList<>();

    // Ofp数据
    private OfpSoapData ofpData;

    // soap数据
    private OfpSoapData soapData;

    // 非甲烷日均值浓度范围, 如果为数组为null, 就是没有数据
    private Float[] nmhcAvgRange;

    // 因子分类日均值
    private List<HistoricalData> factorCategoryDailies = new ArrayList<>();
    private List<Factor> factorGroupDimensions;

    public SiteRangeAvgData(
      String siteName,
      Long siteType,
      Map<Long, Map<String, ReportMonthAvgStatistic>> factorDailyMap,
      Map<String, Factor> factorMap,
      Map<String, FactorGroup> factorCategoryMap,
      Map<String, Object[]> siteFactorMaxValueMap,
      Map<Long, Float[]> nmhcAvgDataMap,
      Map<String, Float> ofpFactorMap,
      Map<String, Float> soapFactorMap
    ) {
      this.siteName = siteName;
      this.siteType = siteType;
      this.statisticData(factorDailyMap, factorMap, factorCategoryMap, siteFactorMaxValueMap, ofpFactorMap, soapFactorMap);
      this.setNMHCAvgRange(nmhcAvgDataMap);
    }

    private SiteRangeAvgData setNMHCAvgRange(Map<Long, Float[]> nmhcAvgDataMap) {
      if (nmhcAvgDataMap == null || nmhcAvgDataMap.size() == 0) {
        return this;
      }

      List<HistoricalData> nmhcAvgDailyDataList = this.nmhcAvgDailyDataList;

      Float[] nmhcAvgRange = new Float[2];
      nmhcAvgDataMap.forEach((ts, nmhcAvgData) -> {
        Float avg = CalculationUtils.digit2(nmhcAvgData[0] / nmhcAvgData[1], 2);
        if (nmhcAvgRange[0] == null) {
          nmhcAvgRange[0] = nmhcAvgRange[1] = avg;
        } else if (avg < nmhcAvgRange[0]) {
          nmhcAvgRange[0] = avg;
        } else if (avg > nmhcAvgRange[1]) {
          nmhcAvgRange[1] = avg;
        }

        HistoricalData historicalData = new HistoricalData();
        historicalData.setFactorId("1026");
        historicalData.setValue(avg);
        historicalData.setDataTime(new Date(ts));
        nmhcAvgDailyDataList.add(historicalData);
      });

      this.nmhcAvgRange = nmhcAvgRange;

      return this;
    }

    private SiteRangeAvgData statisticData(
      Map<Long, Map<String, ReportMonthAvgStatistic>> factorDailyMap,
      Map<String, Factor> factorMap,
      Map<String, FactorGroup> factorCategoryMap,
      Map<String, Object[]> siteFactorMaxValueMap,
      Map<String, Float> ofpFactorMap,
      Map<String, Float> soapFactorMap
    ) {
      if (factorDailyMap.size() == 0) {
        return this;
      }

      List<HistoricalData> factorCategoryDailies = new ArrayList();
      Set<FactorGroup> factorCategorySet = new HashSet();

      Float[] range = new Float[2];
      float[] avgData = new float[]{0, 0};
      Map<String, float[]> factorSumMap = new HashMap();

      factorDailyMap.forEach((ts, dailyAvgStatisticMap) -> {
        if (dailyAvgStatisticMap.size() == 0) {
          return;
        }

        Map<String, Float> factorCategoryDataMap = new HashMap();

        Float[] dailySum = new Float[1];

        dailyAvgStatisticMap.forEach((factorId, dailyAvgStatistic) -> {
          int count = dailyAvgStatistic.getCount();

          if (count == 0) {
            return;
          }

          Factor factor = factorMap.get(factorId);
          float sum = CalculationUtils.UnitConverter(dailyAvgStatistic.getSum(), factor.getStdUnit(), factor.getDisplayUnit(), factor.getMolecular());
          float avg = sum / count;

          dailyAvgStatistic.setSumValue(sum);

          if (dailySum[0] == null) {
            dailySum[0] = 0f;
          }

          dailySum[0] += avg;

          String groupId = factor.getGroupId();

          Float factorCategorySum = factorCategoryDataMap.get(groupId);
          if (factorCategorySum == null) {
            factorCategorySum = 0f;
          }
          factorCategorySum += avg;
          factorCategoryDataMap.put(groupId, factorCategorySum);

          float[] factorSum = factorSumMap.get(factorId);

          if (factorSum == null) {
            factorSumMap.put(factorId, factorSum = new float[]{0, 0});
          }

          factorSum[0] += sum;
          factorSum[1] += count;
        });

        Date dataTime = new Date(ts);
        factorCategoryDataMap.forEach((factorCategoryId, dailyAvg) -> {
          FactorGroup factorGroup = factorCategoryMap.get(factorCategoryId);

          if (factorGroup == null) {
            return;
          }

          HistoricalData historicalData = new HistoricalData();
          historicalData.setDataTime(dataTime);
          historicalData.setFactorId(factorCategoryId);
          historicalData.setValue(dailyAvg);
          factorCategoryDailies.add(historicalData);

          factorCategorySet.add(factorGroup);
        });

        Float sum = dailySum[0];

        if (sum == null) {
          return;
        }

        avgData[0] += sum;
        avgData[1]++;

        if (range[0] == null) {
          range[0] = range[1] = sum;
        } else if (sum < range[0]) {
          range[0] = sum;
        } else if (sum > range[1]) {
          range[1] = sum;
        }
      });

//      if (range != null) {
      this.range = new Float[]{CalculationUtils.digit2(range[0], 2), CalculationUtils.digit2(range[1], 2)};
      this.avg = CalculationUtils.digit2(avgData[0] / avgData[1], 2);
      this.factorCategoryDailies = factorCategoryDailies;
      ArrayList<Factor> factorGroupDimensions = new ArrayList();
      factorCategorySet.forEach(factorCategory -> {
        Factor factor = new Factor();
        factor
          .setId(factorCategory.getId())
          .setDisplayName(factorCategory.getName())
          .setStdUnit("μg/m3")
          .setDisplayUnit("μg/m3");
        factorGroupDimensions.add(factor);
      });
      factorGroupDimensions.sort((o1, o2) -> o1.getId().compareTo(o2.getId()));
      this.factorGroupDimensions = factorGroupDimensions;

      List<PieData> factorAvgList = new ArrayList<>();
      Map<String, PieData> categoryMap = new HashMap<>();

      factorSumMap.forEach((factorId, data) -> {
        Factor factor = factorMap.get(factorId);
        String groupId = factor.getGroupId();
        PieData pieData = categoryMap.get(groupId);

        if (pieData == null) {
          categoryMap.put(groupId, pieData = new PieData(factorCategoryMap.get(groupId).getName()));
        }

        float avg = data[0] / data[1];
        pieData.add(avg);
        factorAvgList.add(new PieData(factorId, factor.getDisplayName(), avg));
      });

      List<PieData> pieDataList = new ArrayList<>(categoryMap.values());
      pieDataList.sort((o1, o2) -> (o2.getValue() - o1.getValue()) > 0 ? 1 : -1);
      this.pieDataList = pieDataList;

      factorAvgList.sort(new Comparator<PieData>() {
        @Override
        public int compare(PieData o1, PieData o2) {
          Float v1 = o1.getValue();
          Float v2 = o2.getValue();
          return (v2 - v1) > 0 ? 1 : -1;
        }
      });

      int factorAvgListSize = Math.min(5, factorAvgList.size());
      List<Object[]> topVocFactorTableData = new ArrayList() {{
        add(new Object[]{"序号", "污染物名称", "月平均浓度<br />（μg/m³）", "最高检出浓度<br />（μg/m³）", "最高浓度检出时间"});
      }};

      for (int i = 0; i < factorAvgListSize; i++) {
        PieData monthAvgData = factorAvgList.get(i);
        Float value = monthAvgData.getValue();

//        if (value < 1.2 && i > 4) {
//          break;
//        }
        Object[] maxValue = siteFactorMaxValueMap.get(monthAvgData.getId());
        topVocFactorTableData.add(new Object[]{i + 1, monthAvgData.getName(), CalculationUtils.digit2(value, 2), CalculationUtils.digit2((Float) maxValue[0], 2), maxValue[1]});
      }

      this.topVocFactorTableData = topVocFactorTableData;
      this.factorAvgList = factorAvgList;
      this.ofpData = new OfpSoapData(
        factorDailyMap,
        new Factor().setId("OFP").setDisplayName("OFP").setStdUnit("μg/m³").setDisplayUnit("μg/m³"),
        factorMap,
        ofpFactorMap,
        factorCategoryMap
      );
      this.soapData = new OfpSoapData(
        factorDailyMap,
        new Factor().setId("SOAp").setDisplayName("SOAp").setStdUnit("μg/m³").setDisplayUnit("μg/m³"),
        factorMap,
        soapFactorMap,
        factorCategoryMap
      );
//      }

      return this;
    }
  }

  @Data
  @ToString
  @Accessors(chain = true)
  public static final class OfpSoapData {
    // 均值
    private double avg;

    private Factor trendDimension;

    // 日均浓度变化柱状堆积图和趋势图用到的维度
    private List<Factor> percentageDimensions;

    // 日均浓度变化柱状堆积图和趋势图列表(图表API: /chart/trend-percentage)
    private List<FactorTimeData> dailyDataList;

    // 排名前10的VOCs因子(图表API: /chart/bar)
    private List<NameValue> topDataList;

    // 分类占比数据(图表API: /chart/pie）
    private List<PieData> categoryPercentageList;

    public OfpSoapData(
      Map<Long, Map<String, ReportMonthAvgStatistic>> dailyFactorStatisticMap,
      Factor trendDimension,
      Map<String, Factor> factorMap,
      Map<String, Float> factorNMap,
      Map<String, FactorGroup> factorCategoryMap
    ) {
      this.trendDimension = trendDimension;

      this.statisticData(
        dailyFactorStatisticMap,
        factorMap,
        factorNMap,
        factorCategoryMap
      );
    }

    private OfpSoapData statisticData(
      Map<Long, Map<String, ReportMonthAvgStatistic>> dailyFactorStatisticMap,
      Map<String, Factor> factorMap,
      Map<String, Float> factorNMap,
      Map<String, FactorGroup> factorCategoryMap
    ) {
      Factor trendDimension = this.trendDimension;
      String trendDimensionId = trendDimension.getId();
      String stdUnit = trendDimension.getStdUnit();
      String displayUnit = trendDimension.getDisplayUnit();

      Map<String, float[]> monthlyFactorStatisticMap = new HashMap<>();
      List<FactorTimeData> dailyDataList = new ArrayList<>();
      Set<String> factorCategoryIdSet = new HashSet<>();

      dailyFactorStatisticMap.forEach((ts, factorStatisticMap) -> {
        if (factorStatisticMap.size() == 0) {
          return;
        }

        Map<String, Float> dailyFactorCategoryAvgMap = new HashMap<>();
        float[] dailySum = new float[]{0};

        factorStatisticMap.forEach((factorId, statistic) -> {
          Float n = factorNMap.get(factorId);
          if (n == null) {
            return;
          }

          int count = statistic.getCount();
          if (count == 0) {
            return;
          }

          float sum = statistic.getSum() * n;
          float avg = sum / count;
          dailySum[0] += avg;

          Factor factor = factorMap.get(factorId);
          String groupId = factor.getGroupId();
          Float factorAvg = dailyFactorCategoryAvgMap.get(groupId);
          if (factorAvg == null) {
            factorAvg = 0F;
          }
          dailyFactorCategoryAvgMap.put(groupId, factorAvg + avg);

          float[] monthlyFactorStatistic = monthlyFactorStatisticMap.get(factorId);
          if (monthlyFactorStatistic == null) {
            monthlyFactorStatisticMap.put(factorId, monthlyFactorStatistic = new float[]{0, 0});
          }
          monthlyFactorStatistic[0] += sum;
          monthlyFactorStatistic[1] += count;
        });

        Date dataTime = new Date(ts);
        dailyDataList.add(new FactorTimeData(dataTime, trendDimensionId, dailySum[0]));

        dailyFactorCategoryAvgMap.forEach((categoryId, value) -> {
          dailyDataList.add(new FactorTimeData(dataTime, categoryId, value));
          factorCategoryIdSet.add(categoryId);
        });
      });

      this.dailyDataList = dailyDataList;

      List<Factor> percentageDimensions = new ArrayList<>();
      factorCategoryIdSet.forEach(factorCategoryId -> {
        FactorGroup factorGroup = factorCategoryMap.get(factorCategoryId);
        Factor factor = new Factor()
          .setId(factorCategoryId)
          .setDisplayName(factorGroup.getName())
          .setStdUnit(stdUnit)
          .setDisplayUnit(displayUnit);

        percentageDimensions.add(factor);
      });
      this.percentageDimensions = percentageDimensions;

      List<NameValue> monthlyFactorAvgList = new ArrayList<>();
      Map<String, Double> monthlyFactorCategoryMap = new HashMap<>();
      double[] sum = new double[]{0};
      monthlyFactorStatisticMap.forEach((factorId, statistic) -> {
        Factor factor = factorMap.get(factorId);
        double avg = statistic[0] / statistic[1];

        monthlyFactorAvgList.add(new NameValue(factor.getDisplayName(), new Float(avg).floatValue()));

        String groupId = factor.getGroupId();
        Double categorySum = monthlyFactorCategoryMap.get(groupId);
        if (categorySum == null) {
          categorySum = 0D;
        }
        monthlyFactorCategoryMap.put(groupId, categorySum + avg);

        sum[0] += avg;
      });

      monthlyFactorAvgList.sort(new Comparator<NameValue>() {
        @Override
        public int compare(NameValue o1, NameValue o2) {
          return o2.getValue() > o1.getValue() ? 1 : -1;
        }
      });

      List<NameValue> topDataList = new ArrayList<>();
      int monthlyFactorAvgListSize = Math.min(monthlyFactorAvgList.size(), 10);
      for (int i = 0; i < monthlyFactorAvgListSize; i++) {
        topDataList.add(monthlyFactorAvgList.get(i));
      }
      this.topDataList = topDataList;

      Double avg = this.avg = CalculationUtils.digit2(sum[0], 2);

      List<PieData> categoryPercentageList = new ArrayList<>();
      monthlyFactorCategoryMap.forEach((factorCategoryId, value) -> {
        categoryPercentageList.add(
          new PieData(
            factorCategoryMap.get(factorCategoryId).getName(),
            CalculationUtils.digit2(value / avg * 100, 2).floatValue()
          )
        );
      });
      categoryPercentageList.sort(new Comparator<PieData>() {
        @Override
        public int compare(PieData o1, PieData o2) {
          return o2.getValue() > o1.getValue() ? 1 : -1;
        }
      });
      this.categoryPercentageList = categoryPercentageList;

      return this;
    }
  }
}
