package more.ssaqms.xiamen.model.report;

import lombok.Data;
import lombok.ToString;
import lombok.experimental.Accessors;
import more.ssaqms.xiamen.model.data.HistoricalData;
import more.ssaqms.xiamen.model.factor.Factor;
import more.ssaqms.xiamen.util.CalculationUtils;

import java.util.*;

/**
 * @author purpose
 * @version 1.0
 * @date 2023/5/9 5:01 下午
 */
@Data
@ToString
@Accessors(chain = true)
public class ReportMonthWeatherLinGangData {

  public static final int[][] WIND_SPEED_LEVELS = new int[][]{
    new int[]{0, 2},
    new int[]{2, 4},
    new int[]{4, 6},
    new int[]{6, 8},
    new int[]{8, 10},
    new int[]{10}
  };

  public static final String[] WIND_DIRECT_CATEGORIES = new String[]{
    "N",
    "NNE",
    "NE",
    "ENE",
    "E",
    "ESE",
    "SE",
    "SSE",
    "S",
    "SSW",
    "SW",
    "WSW",
    "W",
    "WNW",
    "NW",
    "NNW",
  };

  public static final float WIND_DIRECT_INTERVAL = 360 / WIND_DIRECT_CATEGORIES.length;

  private static final float WIND_DIRECT_OFFSET = WIND_DIRECT_INTERVAL / 2;

  // 风速风向因子
  private List<Factor> windFactors;

  // 风速风向因子字典表
  private Map<String, Factor> windFactorMap;

  // 气象因子(不含风速风向因子)
  private List<Factor> factors;

  // 气象因子数据列有(不含风速风向数据)
  private List<LineData> dataList;

  // 风速风向数据列表
  private List<LineData> windDataList;

  // 气温范围
  private Float[] temperatureRange;

  // 主导风向
  private String mainWindDir;

  // 主要风速范围，如果没有数据， 此属性为空
  private int[][] mainWindSpeedRanges;

  // 风速较大的较多风向, 如果没有数据， 此属性为空
  private String biggerMainWindDir;

  // 站房风速风向月数据, key为站房主键
  private Map<String, List<HistoricalData>> siteWindDataListMap;

  public ReportMonthWeatherLinGangData(
    List<HistoricalData> dataList,
    BasicInformation information,
    List<String> factorIds,
    List<String> siteIds
  ) {
    this.setFactors(factorIds, information)
      .setWindData(dataList)
      .setAvgDataMap(dataList, information, factorIds)
      .setWindDescription(dataList);
  }

  private ReportMonthWeatherLinGangData setFactors(List<String> factorIds, BasicInformation information) {
    Map<String, Factor> factorMap = information.getFactorMap();
    List<Factor> factors = new ArrayList<>();
    List<Factor> windFactors = new ArrayList<>();
    Map<String, Factor> windFactorMap = new HashMap();

    factorIds.forEach((factorId) -> {
      switch (factorId) {
        case "1012":
        case "1013":
          Factor windFactor = factorMap.get(factorId);
          windFactors.add(windFactor);
          windFactorMap.put(factorId, windFactor);
          break;
        default:
          factors.add(factorMap.get(factorId));
          break;
      }
    });

    windFactors.sort((o1, o2) -> o1.getId().compareTo(o2.getId()));

    this.factors = factors;
    this.windFactors = windFactors;
    this.windFactorMap = windFactorMap;
    return this;
  }

  private ReportMonthWeatherLinGangData setWindData(List<HistoricalData> dataList) {
    List<HistoricalData> windHistoryDataList = new ArrayList();
    Map<String, Factor> windFactorMap = new HashMap();
    this.windFactors.forEach(windFactor -> {
      windFactorMap.put(windFactor.getId(), windFactor);
    });

    Map<String, List<HistoricalData>> siteWindDataListMap = new HashMap();
    dataList.forEach(data -> {
      if (!windFactorMap.containsKey(data.getFactorId())) {
        return;
      }

      String siteId = data.getSiteId();
      List<HistoricalData> historicalDataList = siteWindDataListMap.get(siteId);
      if (historicalDataList == null) {
        siteWindDataListMap.put(siteId, historicalDataList = new ArrayList());
      }
      historicalDataList.add(data);
    });
    this.siteWindDataListMap = siteWindDataListMap;

    return this;
  }

  private ReportMonthWeatherLinGangData setAvgDataMap(
    List<HistoricalData> dataList,
    BasicInformation information,
    List<String> factorIds
  ) {
    Map<String, Map<Long, Float[]>> statisticMap = this.getStatisticMap(factorIds);
    this.setAvgData(statisticMap, dataList, information);
    return this;
  }

  private Map<String, Map<Long, Float[]>> getStatisticMap(List<String> factorIds) {
    Map<String, Map<Long, Float[]>> statisticMap = new HashMap<>();

    for (String factorId : factorIds) {
      statisticMap.put(factorId, new HashMap<>());
    }

    return statisticMap;
  }

  private void setAvgData(
    Map<String, Map<Long, Float[]>> statisticMap,
    List<HistoricalData> dataList,
    BasicInformation information
  ) {
    Map<Long, Map<String, double[]>> windMap = new LinkedHashMap<>();
    Float[] temperatureRange = new Float[2];

    for (HistoricalData data : dataList) {
      String factorId = data.getFactorId();
      long ts = data.getDataTime().getTime();
      Float value = data.getValue();
      String siteId = data.getSiteId();

      switch (factorId) {
        case "1012":
          Map<String, double[]> siteWindMap = windMap.get(ts);

          if (siteWindMap == null) {
            windMap.put(ts, siteWindMap = new HashMap<>());
          }

          double[] windValue = siteWindMap.get(siteId);

          if (windValue == null) {
            siteWindMap.put(siteId, windValue = new double[2]);
          }

          windValue[0] = value;
          break;
        case "1013":
          siteWindMap = windMap.get(ts);

          if (siteWindMap == null) {
            windMap.put(ts, siteWindMap = new LinkedHashMap<>());
          }

          windValue = siteWindMap.get(siteId);

          if (windValue == null) {
            siteWindMap.put(siteId, windValue = new double[2]);
          }

          windValue[1] = value;
          break;
        default:
          Map<Long, Float[]> tsMap = statisticMap.get(factorId);

          if (tsMap != null) {
            Float[] statistic = tsMap.get(ts);

            if (statistic == null) {
              tsMap.put(ts, statistic = new Float[]{0f, 0f});
            }

            statistic[0] += value;
            statistic[1]++;
          }

          if (factorId.equals("1010")) {
            if (temperatureRange[0] == null) {
              temperatureRange[0] = temperatureRange[1] = value;
            } else if (value < temperatureRange[0]) {
              temperatureRange[0] = value;
            } else if (value > temperatureRange[1]) {
              temperatureRange[1] = value;
            }
          }

          break;
      }
    }

    List<LineData> windDataList = new ArrayList<>();
    List<LineData> formatDataList = new ArrayList<>();
    List<double[]> windList = new ArrayList<>();

    windMap.forEach((ts, siteWindMap) -> {
      List<double[]> siteWindList = new ArrayList<>(siteWindMap.values());
      double[] newWindData = CalculationUtils.calcWsWdVectorAvg(siteWindList);
      Date dataTime = new Date(ts);
      windDataList.add(new LineData("NO_SITE_ID", "1012", dataTime, new Double(newWindData[0]).floatValue()));
      windDataList.add(new LineData("NO_SITE_ID", "1013", dataTime, new Double(newWindData[1]).floatValue()));
      windList.addAll(siteWindList);

      statisticMap.forEach((factorId, tsMap) -> {
        Float[] statisticData = tsMap.get(ts);

        if (statisticData == null) {
          return;
        }

        formatDataList.add(new LineData("NO_SITE_ID", factorId, dataTime, statisticData[0] / statisticData[1]));
      });
    });

    this.mainWindDir = CalculationUtils.getWdString(CalculationUtils.calcWsWdVectorAvg(windList)[1]);
    this.temperatureRange = temperatureRange;
    this.windDataList = windDataList;
    this.dataList = formatDataList;
  }


  private ReportMonthWeatherLinGangData setWindDescription(List<HistoricalData> dataList) {
    int windSpeedLevelsLen = WIND_SPEED_LEVELS.length;
    Map<String, Map<Long, Float>> siteWindDirectDataMap = new HashMap();
    List<HistoricalData> filterWindSpeedDataList = new ArrayList();
    List<HistoricalData>[] categoryWindSpeedDataList = new ArrayList[windSpeedLevelsLen];
    dataList.forEach(data -> {
      switch (data.getFactorId()) {
        case "1012":
          Float value = data.getValue();
          int windSpeedLevelIndex = (int) Math.min(windSpeedLevelsLen - 1, Math.floor(value / 2));
          List<HistoricalData> historicalDataList = categoryWindSpeedDataList[windSpeedLevelIndex];
          if (historicalDataList == null) {
            historicalDataList = categoryWindSpeedDataList[windSpeedLevelIndex] = new ArrayList();
          }
          historicalDataList.add(data);
          // 过滤大于10米/秒的风速数据
          if (value > 10) {
            filterWindSpeedDataList.add(data);
          }
          break;
        case "1013":
          String siteId = data.getSiteId();
          Map<Long, Float> windDirectDataMap = siteWindDirectDataMap.get(siteId);
          if (windDirectDataMap == null) {
            siteWindDirectDataMap.put(siteId, windDirectDataMap = new HashMap());
          }
          windDirectDataMap.put(data.getDataTime().getTime(), data.getValue());
          break;
      }
    });

    int maxTimes = 0;
    List<Integer> mainWindSpeedIndexes = new ArrayList();
    for (int i = 0; i < windSpeedLevelsLen; i++) {
      List<HistoricalData> historicalDataList = categoryWindSpeedDataList[i];
      if (historicalDataList == null) {
        continue;
      }

      int size = historicalDataList.size();
      if (size == maxTimes) {
        mainWindSpeedIndexes.add(i);
      } else if (size > maxTimes) {
        maxTimes = size;
        mainWindSpeedIndexes = new ArrayList();
        mainWindSpeedIndexes.add(i);
      }
    }

    int mainWindSpeedIndexesLen = mainWindSpeedIndexes.size();
    if (mainWindSpeedIndexesLen > 0) {
      int[][] mainWindSpeedRanges = new int[mainWindSpeedIndexesLen][2];
      for (int i = 0; i < mainWindSpeedIndexesLen; i++) {
        mainWindSpeedRanges[i] = WIND_SPEED_LEVELS[mainWindSpeedIndexes.get(i)];
      }
      this.mainWindSpeedRanges = mainWindSpeedRanges;
    }


    Map<Long, Map<String, double[]>> windMap = new LinkedHashMap<>();
    filterWindSpeedDataList.forEach(data -> {
      String siteId = data.getSiteId();
      Map<Long, Float> windDirectDataMap = siteWindDirectDataMap.get(siteId);
      if (windDirectDataMap == null) {
        return;
      }

      long ts = data.getDataTime().getTime();
      Float windDirectValue = windDirectDataMap.get(ts);
      if (windDirectValue == null) {
        return;
      }

      Map<String, double[]> siteWindMap = windMap.get(siteId);
      if (siteWindMap == null) {
        windMap.put(ts, siteWindMap = new HashMap());
      }

      siteWindMap.put(siteId, new double[]{data.getValue(), windDirectValue});
    });

    List<double[]> windList = new ArrayList<>();
    windMap.forEach((ts, siteWindMap) -> {
      List<double[]> siteWindList = new ArrayList<>(siteWindMap.values());
      double[] newWindData = CalculationUtils.calcWsWdVectorAvg(siteWindList);
      windList.addAll(siteWindList);
    });

    this.biggerMainWindDir = CalculationUtils.getWdString(CalculationUtils.calcWsWdVectorAvg(windList)[1]);


    return this;
  }

  private int getWindDirectCategoryIndex(float angle) {
    return (int) Math.floor(((angle + WIND_DIRECT_OFFSET) % 360) / WIND_DIRECT_INTERVAL);
  }
}
